import inspect
import os
from pathlib import PurePath
from typing import List, Dict

from docutils.readers import doctree
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.jinja2glue import SphinxFileSystemLoader

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent
from kaybee.plugins.resources.action import ResourceAction
from kaybee.plugins.resources.container import ResourcesContainer
from kaybee.plugins.resources.directive import ResourceDirective
from kaybee.utils.rst import get_rst_title


@kb.event(SphinxEvent.BI, scope='resources')
def handle_builderinited(kb_app: kb, sphinx_app: Sphinx):
    pass


@kb.event(SphinxEvent.EBRD, scope='resource', system_order=40)
def initialize_resources_container(kb_app: kb,
                                   sphinx_app: Sphinx,
                                   sphinx_env: BuildEnvironment,
                                   docnames=List[str],
                                   ):
    if not hasattr(sphinx_app.env, 'resources'):
        sphinx_app.env.resources = ResourcesContainer()


@kb.event(SphinxEvent.EBRD, scope='resource', system_order=50)
def register_template_directory(kb_app: kb,
                                sphinx_app: Sphinx,
                                sphinx_env: BuildEnvironment,
                                docnames=List[str],
                                ):
    """ Add this resource's templates dir to template paths """

    template_bridge = sphinx_app.builder.templates

    actions = ResourceAction.get_callbacks(kb_app)

    for action in actions:
        f = os.path.dirname(inspect.getfile(action))
        template_bridge.loaders.append(SphinxFileSystemLoader(f))


@kb.event(SphinxEvent.EBRD, scope='resource', system_order=55)
def add_directives(kb_app: kb,
                   sphinx_app: Sphinx,
                   sphinx_env: BuildEnvironment,
                   docnames=List[str],
                   ):
    """ For each resource type, register a new Sphinx directive """

    for k, v in list(kb_app.config.resources.items()):
        sphinx_app.add_directive(k, ResourceDirective)


@kb.event(SphinxEvent.EU, scope='resource')
def process_field_handlers(kb_app: kb,
                           sphinx_app: Sphinx,
                           sphinx_env: BuildEnvironment
                           ):
    resources = sphinx_app.env.resources
    for resource in resources.values():
        images = getattr(resource.props, 'images', None)
        if images:
            for prop in images:
                t = resource.props.fields['images'].type_
                if hasattr(prop, 'env_updated'):
                    prop.env_updated(
                        kb_app,
                        sphinx_app,
                        sphinx_env,
                        resource
                    )


@kb.event(SphinxEvent.DREAD, scope='resource')
def stamp_title(kb_app: kb,
                sphinx_app: Sphinx,
                doctree: doctree):
    """ Walk the tree and extra RST title into resource.title """

    # First, find out which resource this is. Won't be easy.
    resources = sphinx_app.env.resources
    confdir = sphinx_app.confdir
    source = PurePath(doctree.attributes['source'])

    # Get the relative path inside the docs dir, without .rst, then
    # get the resource
    docname = str(source.relative_to(confdir)).split('.rst')[0]
    resource = resources.get(docname)

    if resource:
        # Stamp the title on the resource
        title = get_rst_title(doctree)
        resource.title = title


@kb.event(SphinxEvent.HPC, scope='resource')
def resource_into_html_context(
        kb_app: kb,
        sphinx_app: Sphinx,
        pagename,
        templatename: str,
        context,
        doctree: doctree) -> Dict[str, str]:
    # Always put the resources db into Jinja2 context
    resources = sphinx_app.env.resources
    context['resources'] = resources

    # Get the resource for this pagename. If no match, then this pagename
    # must be a genericpage
    resource = resources.get(pagename)
    if resource:
        context['resource'] = resource
        templatename = resource.template(resources) + '.html'
        return dict(templatename=templatename)


@kb.dumper('resources')
def dump_settings(kb_app: kb, sphinx_env: BuildEnvironment):
    # First get the kb app configuration for resources
    config = {
        k: v.__module__ + '.' + v.__name__
        for (k, v) in kb_app.config.resources.items()
    }

    # Next, get the actual resources in the app.resources DB
    resources = sphinx_env.resources
    values = {k: v.__json__(resources) for (k, v) in resources.items()}
    resources = dict(
        config=config,
        values=values
    )
    return dict(resources=resources)



"""

Intercept and re-render Sphinx built-in toctrees.

This class renders alternate HTML for the existing Sphinx toctree.
You use toctree as normal, then during Sphinx processing events, we find
toctrees and re-render them with our HTML.

Why not make our own directive? toctrees are complex. Working with the
rendering of the existing version is easier.

"""

from typing import List, Tuple

from sphinx.application import Sphinx
from sphinx.util.osutil import relative_uri


class BaseToctree:
    """ Basis for a subclass with a template """
    entries: List = []
    result_count = 0
    template = 'toctree'

    def __init__(self, docname: str):
        self.docname = docname

    def set_entries(self, entries: List[Tuple[str, str]], titles, resources):
        """ Provide the template the data for the toc entries """

        self.entries = []
        for flag, pagename in entries:
            title = titles[pagename].children[0]
            resource = resources.get(pagename, None)
            if resource and hasattr(resource,
                                    'is_published') and not \
                    resource.is_published:
                continue
            # Even if there is no resource for this tocentry, we can
            # use the toctree info
            self.entries.append(dict(
                title=title, href=pagename, resource=resource
            ))

        self.result_count = len(self.entries)

    def pathto_docname(self, target_docname):
        """ Mimic the sphinx.builders.html.pathto inline closure.

        We need a way to resolve relative URLs so hrefs work when a Sphinx
        site is served in a subdirectory. Sphinx has an odd inline function
        called pathto that has some local variables. This function serves the
        purpose for paths to docs as well as static resourcees. Let's split
        that and have our own, for widgets which aren't using the Sphinx
        Jinja2 renderer.

        """

        uri = relative_uri(self.docname, target_docname)
        return uri

    def render(self, builder, context, sphinx_app: Sphinx):
        """ Given a Sphinx builder and context with site in it,
         generate HTML """

        context['sphinx_app'] = sphinx_app
        context['toctree'] = self

        html = builder.templates.render(self.template + '.html', context)
        return html



"""

An out-of-the-box resource type.

This can be used in RST with something like::

   .. resource::
      template: sometemplate.html

"""

from kaybee.app import kb

from kaybee.plugins.resources.base_resource import BaseResource


@kb.resource('resource')
class Resource(BaseResource):
    pass



"""

Widget providing a number of queries to be performed then laid out
in columns.

"""
from typing import List

from pydantic import BaseModel

from kaybee.app import kb
from kaybee.plugins.queries.props_model import BaseQueryModel
from kaybee.plugins.widgets.base_widget import (
    BaseWidget,
    BaseWidgetModel,
)
from kaybee.plugins.queries.service import Query


class QuerySectionModel(BaseModel):
    label: str
    style: str = None
    query: BaseQueryModel


class QuerylistModel(BaseWidgetModel):
    queries: List[QuerySectionModel]


@kb.widget('querylist')
class QuerylistWidget(BaseWidget):
    props: QuerylistModel

    def make_context(self, context, sphinx_app):
        result_sets = []
        for query_section in self.props.queries:
            result_set = dict(
                label=query_section.label,
                style=query_section.style,
            )
            query = query_section.query
            results = Query.filter_collection(
                sphinx_app.env.resources,
                rtype=query.rtype,
                parent_name=query.parent_name,
                sort_value=query.sort_value,
                limit=query.limit,
                reverse=query.reverse
            )
            result_set['results'] = results
            result_sets.append(result_set)

        context['result_sets'] = result_sets



from operator import attrgetter
from typing import List, Dict, Optional

import pydash
from pydash import py_


class Query:
    def __init__(self, docname):
        self.docname = docname

    @classmethod
    def _attr_lambda(cls, key, value):
        # For use in filter expressions. If the value is None, then
        # don't do any filtering. If the value is not None, then we need
        # to filter.
        if value is None:
            return lambda x: True
        else:
            return lambda x: getattr(x, key, None) == value

    @classmethod
    def _prop_lambda(cls, key, value):
        # For use in filter expressions. If the value is None, then
        # don't do any filtering. If the value is not None, then we need
        # to filter.
        if value is None:
            return lambda x: True
        else:
            return lambda x: getattr(x.props, key, None) == value

    @classmethod
    def _filter_parents(cls, collection, parent_name):
        if parent_name is None:
            return lambda x: True
        else:
            return lambda x: collection[parent_name] in x.parents(collection)

    @classmethod
    def _sort_key_lamda(cls, sort_value):
        if sort_value is None:
            # The dummy user explicity said to sort on nothing. What to
            # sort on then, when they don't want the default sort_value of
            # 'title' ? For now, docname
            return attrgetter('docname')
        elif sort_value == 'title':
            return attrgetter('title')
        else:
            def get_prop(resource):
                prop = getattr(resource.props, sort_value, None)
                if prop is None:
                    return ''
                else:
                    return prop
            return get_prop

    @classmethod
    def filter_collection(self,
                          collection,
                          rtype: str = None,
                          sort_value: Optional[str] = 'title',
                          reverse: bool = False,
                          limit: int = None,
                          parent_name: str = None,
                          props: List[Dict[str, str]] = [],
                          ):

        # Set the limit
        if limit is None:
            limit = len(collection.values())

        # Filter those results based on arbitrary key-value pairs
        r1 = collection
        for prop in props:
            r1 = pydash.filter_(r1,
                                Query._prop_lambda(prop['key'], prop['value']))

        r1 = py_(r1) \
            .filter_(Query._attr_lambda('rtype', rtype)) \
            .filter_(Query._filter_parents(collection, parent_name)) \
            .sort_by(Query._sort_key_lamda(sort_value),
                     reverse=reverse
                     ) \
            .slice(0, limit)

        return r1.value()



"""

An out-of-the-box article type.

This can be used in RST with something like::

   .. article::
      template: sometemplate.html

"""

from kaybee.app import kb

from kaybee.plugins.articles.base_article import BaseArticle, BaseArticleModel


@kb.resource('article')
class Article(BaseArticle):
    props: BaseArticleModel



"""

An action for the kb registry app.

"""

import dectate


class ResourceAction(dectate.Action):
    config = {
        'resources': dict
    }

    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def identifier(self, resources):
        return self.name

    # noinspection PyMethodOverriding
    def perform(self, obj, resources):
        resources[self.name] = obj

    @classmethod
    def get_callbacks(cls, registry):
        # Presumes the registry has been committed

        q = dectate.Query('resource')
        return [args[1] for args in q(registry)]



import inspect
import os
from typing import List

from docutils import nodes
from docutils.readers import doctree
from sphinx.application import Sphinx
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment import BuildEnvironment
from sphinx.jinja2glue import SphinxFileSystemLoader

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent
from kaybee.plugins.widgets.action import WidgetAction
from kaybee.plugins.widgets.node import widget
from kaybee.plugins.widgets.directive import WidgetDirective


@kb.event(SphinxEvent.BI, scope='widgets')
def add_widget_node(kb_app: kb, sphinx_app: Sphinx):
    sphinx_app.add_node(widget)


@kb.event(SphinxEvent.EBRD, scope='widgets', system_order=40)
def initialize_widgets_container(kb_app: kb,
                                 sphinx_app: Sphinx,
                                 sphinx_env: BuildEnvironment,
                                 docnames=List[str],
                                 ):
    if not hasattr(sphinx_app.env, 'widgets'):
        sphinx_app.env.widgets = dict()


@kb.event(SphinxEvent.EBRD, scope='widgets', system_order=50)
def register_template_directory(kb_app: kb,
                                sphinx_app: Sphinx,
                                sphinx_env: BuildEnvironment,
                                docnames=List[str],
                                ):
    """ Add this widget's templates dir to template paths """

    template_bridge = sphinx_app.builder.templates

    actions = WidgetAction.get_callbacks(kb_app)

    for action in actions:
        f = os.path.dirname(inspect.getfile(action))
        template_bridge.loaders.append(SphinxFileSystemLoader(f))


@kb.event(SphinxEvent.EBRD, scope='widgets', system_order=60)
def register_widget_directive(kb_app: kb,
                              sphinx_app: Sphinx,
                              sphinx_env: BuildEnvironment,
                              docnames=List[str],
                              ):
    # Register a directive
    for k, v in list(kb_app.config.widgets.items()):
        sphinx_app.add_directive(k, WidgetDirective)


@kb.event(SphinxEvent.DRES, scope='widgets')
def render_widgets(kb_app: kb,
                   sphinx_app: Sphinx,
                   doctree: doctree,
                   fromdocname: str,
                   ):
    """ Go through docs and replace widget directive with rendering """

    builder: StandaloneHTMLBuilder = sphinx_app.builder

    for node in doctree.traverse(widget):
        # Render the output
        w = sphinx_app.env.widgets.get(node.name)
        context = builder.globalcontext.copy()

        # Add in certain globals
        context['resources'] = sphinx_app.env.resources
        context['references'] = sphinx_app.env.references
        output = w.render(sphinx_app, context)

        # Put the output into the node contents
        listing = [nodes.raw('', output, format='html')]
        node.replace_self(listing)


@kb.dumper('widgets')
def dump_settings(kb_app: kb, sphinx_env: BuildEnvironment):
    # First get the kb app configuration for widgets
    config = {
        k: v.__module__ + '.' + v.__name__
        for (k, v) in kb_app.config.widgets.items()
    }

    # Next, get the actual widgets in the app.widgets DB
    widgets = sphinx_env.widgets
    values = {k: v.__json__() for (k, v) in widgets.items()}

    widgets = dict(
        config=config,
        values=values
    )
    return dict(widgets=widgets)



import pytest

pytestmark = pytest.mark.sphinx('html', testroot='references1')


@pytest.mark.parametrize('page', ['index.html', ], indirect=True)
class TestReferences1:

    def test_index(self, page):
        h1 = page.find('h1').contents[0].strip()
        assert 'Indexpage' == h1
        li = page.find('li').contents[0].strip()
        assert 'reference1' == li

        # Let's look at the two types of generated reference links
        refs = page.find_all("a", class_="internal")
        assert 4 == len(refs)  # Includes toctree on
        derived_title = refs[0].find('em').contents[0].strip()
        assert 'Reference 1' == derived_title
        explicit_title = refs[1].find('em').contents[0].strip()
        assert 'Reference 1-ish' == explicit_title


@pytest.mark.parametrize('json_page', ['debug_dump.json', ], indirect=True)
class TestReferences1Debug:

    def test_references(self, json_page):
        assert 'references' in json_page
        references = json_page['references']

        # First: The reference reference
        reference = references['values']['reference']
        assert 'reference1' in reference
        assert 'reference1' == reference['reference1']['docname']

        # Let's see on the other side with the source of the reference
        reference1 = json_page['resources']['values']['index']
        assert 'reference1' == reference1['props']['references']['reference'][
            0]

        # Next: The author reference
        author_targets = references['values']['author']
        assert 'author1' in author_targets
        assert 'author1' == author_targets['author1']['docname']

        # Let's see on the other side with the source of the reference
        author1 = json_page['resources']['values']['index']
        assert 'author1' == author1['props']['references']['author'][0]


@pytest.mark.parametrize('json_page', ['debug_dump.json', ], indirect=True)
class TestReferences1RegistryDebug:

    def test_references(self, json_page):
        assert 'references' in json_page
        references = json_page['references']
        assert 'config' in references

        # references1 contains one registered handler
        config = references['config']
        v = config['reference']
        assert 'kaybee.plugins.references.reference.Reference' == v

        # one value in references
        values = references['values']
        assert 4 == len(values)



from kaybee.app import kb
from kaybee.plugins.widgets.base_widget import (
    BaseWidget,
    BaseWidgetModel,
)


class SectionboxModel(BaseWidgetModel):
    heading: str = None
    subheading: str = None
    style: str = None


@kb.widget('sectionbox')
class SectionboxWidget(BaseWidget):
    props: SectionboxModel



"""
Miscellaneous functions used in Kaybee.
"""
from typing import Optional, Any

from docutils import nodes
from docutils.core import publish_parts
from docutils.frontend import OptionParser
from docutils.nodes import Node, document, paragraph
from docutils.parsers.rst import Parser
from docutils.utils import new_document


def rst_document(rst_string: str) -> document:
    default_settings = OptionParser(
        components=(Parser,)).get_default_values()
    document = new_document(rst_string, default_settings)
    parser = Parser()
    parser.parse(rst_string, document)
    return document


def rst_to_html(input_string: str) -> str:
    """ Given a string of RST, use docutils to generate html """

    overrides = dict(input_encoding='unicode', doctitle_xform=True,
                     initial_header_level=1)
    parts = publish_parts(
        writer_name='html',
        source=input_string,
        settings_overrides=overrides
    )
    return parts['html_body']


def get_rst_title(rst_doc: Node) -> Optional[Any]:
    """ Given some RST, extract what docutils thinks is the title """

    for title in rst_doc.traverse(nodes.title):
        return title.astext()

    return None


def get_rst_excerpt(rst_doc: document, paragraphs: int = 1) -> str:
    """ Given rst, parse and return a portion """

    texts = []
    for count, p in enumerate(rst_doc.traverse(paragraph)):
        texts.append(p.astext())
        if count + 1 == paragraphs:
            break
    return ' '.join(texts)



from pydantic import BaseModel


class ArticlesModel(BaseModel):
    use_toctree: bool = False
    datefmt_short: str = '%b %d'
    datefmt_long: str = '%Y/%m/%d'
    datefmt_full: str = '%Y/%m/%d %H:%M'

    class Config:
        ignore_extra = False



from typing import List

from docutils.parsers.rst import Directive

from kaybee import app
from kaybee.plugins.widgets.node import widget


class WidgetDirective(Directive):
    has_content = True

    @classmethod
    def get_widget_class(cls, widget_directive):
        """ Make this easy to mock """
        return app.kb.config.widgets[widget_directive]

    @classmethod
    def split_content(cls, content: List[str]):
        # The directive gives us the RST body in this directive as
        # a list of strings, one for each line, with empty strings for
        # blank lines. Start by re-assembling one big string (would be
        # better to do some iterator-thingy.)
        content = '\n'.join(content)

        # First strip any leading/trailing blank lines from indented content
        content = content.strip()

        # Split the YAML from the rest of the body, if any blank
        # lines in the middle
        split_content = content.split('\n\n')
        yaml_content = split_content[0]

        # Re-assemble the RST content in the widget body, if any
        if len(split_content) > 1:
            rst_content = '\n\n'.join(split_content[1:])
        else:
            rst_content = None

        return yaml_content, rst_content

    def get_widget(self, docname):
        # Get the info from this directive and make instance
        wtype = self.name

        yaml_content, rst_content = self.split_content(self.content)
        widget_content = '\n'.join(self.content)
        widget_class = WidgetDirective.get_widget_class(wtype)
        return widget_class(docname, wtype, yaml_content, rst_content)

    @property
    def docname(self):
        return self.state.document.settings.env.docname

    @property
    def widgets(self):
        return self.state.document.settings.env.widgets

    def run(self):
        """ Run at parse time.

        When the documents are initially being scanned, this method runs
        and does two things: (a) creates an instance that is added to
        the site's widgets, and (b) leaves behind a placeholder docutils
        node that can later be processed after the docs are resolved.
        The latter needs enough information to retrieve the former.

        """

        this_widget = self.get_widget(self.docname)

        self.widgets[repr(this_widget)] = this_widget

        # Now add the node to the doctree
        widget_node = widget()
        ids = [repr(this_widget)]
        names = [self.name]
        attrs = dict(ids=ids, names=names)
        widget_node.update_basic_atts(attrs)
        return [widget_node]



"""

An out-of-the-box toctree type.

"""

from kaybee.app import kb

from kaybee.plugins.articles.base_toctree import BaseToctree

# When no context argument is supplied, it sets the default
# toctree handler


@kb.toctree(system_order=80)
class Toctree(BaseToctree):
    pass



"""

Widget providing a number of queries to be performed then laid out
in columns.

"""
from typing import Dict, List

from pydantic import BaseModel

from kaybee.app import kb
from kaybee.plugins.queries.props_model import BaseQueryModel
from kaybee.plugins.queries.service import Query
from kaybee.plugins.widgets.base_widget import (
    BaseWidget,
    BaseWidgetModel,
)


class FeatureTile(BaseModel):
    style: str = ''
    heading: str
    subheading: str = None
    bullets: List[str] = None
    more_href: str = None


class FeaturetilesModel(BaseWidgetModel):
    rows: List[List[FeatureTile]] = []


@kb.widget('featuretiles')
class FeaturetilesWidget(BaseWidget):
    props: FeaturetilesModel



"""

Actions for the kb registry app.

"""
from typing import Optional

import dectate


class ToctreeAction(dectate.Action):
    config = {
        'toctrees': dict
    }

    def __init__(self, context: Optional[str] = None,
                 system_order: Optional[int] = 80):
        super().__init__()
        self.context = context
        self.system_order = system_order

    def identifier(self, toctrees):
        return f'{self.context}-{self.system_order}'

    # noinspection PyMethodOverriding
    def perform(self, obj, toctrees):
        name = f'{self.context}-{self.system_order}'
        toctrees[name] = obj

    @classmethod
    def get_callbacks(cls, registry):
        # Presumes the registry has been committed

        q = dectate.Query('toctree')
        return [args[1] for args in q(registry)]

    @classmethod
    def get_for_context(cls, registry, context=None):
        q = dectate.Query('toctree')
        # Get the registrations for this context, sorted by
        # system_order
        qr = sorted(
            q.filter(context=context)(registry),
            key=lambda t: t[0].system_order
        )
        # Return the first match's decorated object
        return qr[0][1]



from kaybee.app import kb

from kaybee.plugins.widgets.base_widget import (
    BaseWidget,
    BaseWidgetModel,
)


class VideoPlayerModel(BaseWidgetModel):
    width: int = 640
    height: int = 360
    src: str
    frameborder: int = 0
    allowfullscreen: bool = True


@kb.widget('videoplayer')
class VideoPlayer(BaseWidget):
    props: VideoPlayerModel



"""

An out-of-the-box section type.

This can be used in RST with something like::

   .. section::
      template: section.html

"""

from kaybee.app import kb

from kaybee.plugins.articles.base_section import BaseSection


@kb.resource('section')
class Section(BaseSection):

    def featured_resource(self, resources):
        fd = self.props.featured_resource
        if not fd:
            return None
        else:
            return resources[fd]



class Genericpage:
    def __init__(self, docname):
        self.docname = docname

    def __repr__(self):
        # Primarily useful in pytest parameters
        return self.docname

    @classmethod
    def template(cls, resources):
        # By default, just use the Sphinx built-in template for pages
        templatename = 'page'

        root = resources.get('index', False)
        if root:
            acquireds = getattr(root.props, 'acquireds', False)
            if acquireds:
                # First try in the per-type acquireds
                gp_acquireds = acquireds.get('genericpage')
                if gp_acquireds:
                    prop_acquired = gp_acquireds.get('template')
                    if prop_acquired:
                        return prop_acquired

                # Next in the "all" section of acquireds
                all_acquireds = acquireds.get('all')
                if all_acquireds:
                    prop_acquired = all_acquireds.get('template')
                    if prop_acquired:
                        return prop_acquired

        return templatename

    def __json__(self, resources):
        return dict(
            docname=self.docname,
            template=self.template(resources),
            repr=repr(self),
        )



from pathlib import PurePath
from typing import List, Dict

from docutils.readers import doctree
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent
from kaybee.plugins.genericpage.action import GenericpageAction


@kb.event(SphinxEvent.EBRD, scope='genericpage', system_order=40)
def initialize_genericpages_container(kb_app: kb,
                                      sphinx_app: Sphinx,
                                      sphinx_env: BuildEnvironment,
                                      docnames=List[str],
                                      ):
    if not hasattr(sphinx_app.env, 'genericpages'):
        sphinx_app.env.genericpages = dict()


@kb.event(SphinxEvent.DREAD, scope='genericpage')
def add_genericpage(kb_app: kb, sphinx_app: Sphinx,
                    doctree: doctree):
    resources = sphinx_app.env.resources

    confdir = sphinx_app.confdir
    source = PurePath(doctree.attributes['source'])

    # Get the relative path inside the docs dir, without .rst, then
    # get the resource
    docname = str(source.relative_to(confdir)).split('.rst')[0]

    resource = resources.get(docname)
    if resource is None:
        # Couldn't find a resource for this document, so let's make it
        # a genericpage
        gp_class = GenericpageAction.get_genericpage(kb_app)
        gp = gp_class(docname)
        sphinx_app.env.genericpages[docname] = gp
        return gp


@kb.event(SphinxEvent.HPC, scope='genericpage')
def genericpage_into_html_context(
        kb_app: kb,
        genericpages_sphinx_app: Sphinx,
        pagename,
        templatename: str,
        context,
        doctree: doctree) -> Dict[str, str]:

    # Get the resource for this pagename. If no match, then this pagename
    # must be a genericpage
    resources = genericpages_sphinx_app.env.resources
    resource = resources.get(pagename)
    if not resource:
        genericpages = genericpages_sphinx_app.env.genericpages
        gp = genericpages.get(pagename)
        if gp:
            context['genericpage'] = gp
            context['resources'] = resources
            template = gp.template(resources)
            templatename = template + '.html'
            return dict(templatename=templatename)


@kb.dumper('genericpages')
def dump_settings(kb_app: kb, sphinx_env: BuildEnvironment):
    config = {
        k: v.__module__ + '.' + v.__name__
        for (k, v) in kb_app.config.genericpages.items()
    }

    # Get the Genericpage values in the app
    values = {}
    resources = sphinx_env.resources
    gps = sphinx_env.genericpages
    values = {k: v.__json__(resources) for (k, v) in gps.items()}

    genericpages = dict(
        config=config,
        values=values
    )
    return dict(genericpages=genericpages)



import os
from typing import List

from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.jinja2glue import SphinxFileSystemLoader

from kaybee.app import kb

from kaybee.plugins.events import SphinxEvent


@kb.event(SphinxEvent.EBRD, scope='localtemplates')
def handle_beforereaddocs(kb_app: kb, sphinx_app: Sphinx,
                          sphinx_env: BuildEnvironment,
                          docnames: List[str]):
    confdir = os.path.join(sphinx_app.confdir, '_templates')
    template_bridge = sphinx_app.builder.templates
    template_bridge.loaders.insert(0, SphinxFileSystemLoader(confdir))



from kaybee.plugins.articles.base_article import (
    BaseArticle,
    BaseArticleModel,
)


class BaseSectionModel(BaseArticleModel):
    featured_resource: str = None  # docname for this section's feature
    subheading: str = None


class BaseSection(BaseArticle):
    props: BaseSectionModel

    def get_featured_resource(self, resources):
        fr = self.props.featured_resource
        if not fr:
            return None
        else:
            return resources[fr]

    def __json__(self, resources):
        d = super().__json__(resources)
        r = self.get_featured_resource(resources)
        if r:
            d['get_featured_resource'] = r.docname

        return d



from pydantic import BaseModel

from kaybee.plugins.articles.settings import ArticlesModel
from kaybee.plugins.debugdumper.settings import DebugdumperModel


class KaybeeSettings(BaseModel):
    plugins_dir: str = 'kaybee_plugins'
    articles: ArticlesModel = ArticlesModel()
    debugdumper: DebugdumperModel = DebugdumperModel()

    class Config:
        ignore_extra = False



"""

Jinja2 filters

The articles system needs some rich ways to format things. For
example, different ways to show dates. We'll use Jinja2 filters for
this.

The filters are registered as instance callables. This lets us put
psuedo-globals such as sphinx_app into the scope and not have to
pass that in from the template usage.

"""

from sphinx.application import Sphinx


class BaseFilter:
    def __init__(self, sphinx_app: Sphinx):
        self.sphinx_app = sphinx_app


class DatetimeFilter(BaseFilter):
    def __call__(self, date_value, fmt='long'):
        if date_value is None:
            # Some resource props values are None
            return

        # Get the setting from the articles setting
        settings = self.sphinx_app.config.kaybee_settings.articles
        fmt_setting = getattr(settings, 'datefmt_' + fmt)

        # Return value
        response = date_value.strftime(fmt_setting)
        return response



import inspect
import os
from operator import attrgetter
from pathlib import PurePath
from typing import List, Dict

from docutils import nodes
from docutils.readers import doctree
from sphinx.addnodes import toctree
from sphinx.application import Sphinx
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment import BuildEnvironment
from sphinx.jinja2glue import SphinxFileSystemLoader

from kaybee.app import kb
from kaybee.plugins.articles.actions import ToctreeAction
from kaybee.plugins.articles.filters import DatetimeFilter
from kaybee.plugins.events import SphinxEvent
from kaybee.plugins.settings.model import KaybeeSettings
from kaybee.utils.rst import get_rst_excerpt


@kb.event(SphinxEvent.BI, scope='articles')
def register_filters(kb_app: kb, sphinx_app: Sphinx):
    dtf = DatetimeFilter(sphinx_app)
    filters = sphinx_app.builder.templates.environment.filters
    filters['datetime_fmt'] = dtf


@kb.event(SphinxEvent.EBRD, scope='toctrees')
def register_template_directory(kb_app: kb,
                                sphinx_app: Sphinx,
                                sphinx_env: BuildEnvironment,
                                docnames=List[str],
                                ):
    """ Add this widget's templates dir to template paths """

    template_bridge = sphinx_app.builder.templates

    actions = ToctreeAction.get_callbacks(kb_app)

    for action in actions:
        f = os.path.dirname(inspect.getfile(action))
        template_bridge.loaders.append(SphinxFileSystemLoader(f))


@kb.event(SphinxEvent.DRES, scope='toctrees')
def render_toctrees(kb_app: kb, sphinx_app: Sphinx, doctree: doctree,
                    fromdocname: str):
    """ Look in doctrees for toctree and replace with custom render """

    # Only do any of this if toctree support is turned on in KaybeeSettings.
    # By default, this is off.
    settings: KaybeeSettings = sphinx_app.config.kaybee_settings
    if not settings.articles.use_toctree:
        return

    # Setup a template and context
    builder: StandaloneHTMLBuilder = sphinx_app.builder
    env: BuildEnvironment = sphinx_app.env

    # Toctree support. First, get the registered toctree class, if any
    registered_toctree = ToctreeAction.get_for_context(kb_app)
    for node in doctree.traverse(toctree):
        if node.attributes['hidden']:
            continue
        custom_toctree = registered_toctree(fromdocname)
        context = builder.globalcontext.copy()
        context['sphinx_app'] = sphinx_app

        # Get the toctree entries. We only handle one level of depth for
        # now. To go further, we need to recurse like sphinx's
        # adapters.toctree._toctree_add_classes function
        entries = node.attributes['entries']

        # The challenge here is that some items in a toctree
        # might not be resources in our "database". So we have
        # to ask Sphinx to get us the titles.
        custom_toctree.set_entries(entries, env.titles,
                                   sphinx_app.env.resources)
        output = custom_toctree.render(builder, context, sphinx_app)

        # Put the output into the node contents
        listing = [nodes.raw('', output, format='html')]
        node.replace_self(listing)


@kb.event(SphinxEvent.DREAD, scope='toctrees')
def resource_toctrees(kb_app: kb,
                      sphinx_app: Sphinx,
                      doctree: doctree):
    # Find any toctrees in doc (at most one, hopefully) and record
    # onto the resource state

    # First, find out which resource this is. Won't be easy.
    resources = sphinx_app.env.resources
    confdir = sphinx_app.confdir
    source = PurePath(doctree.attributes['source'])

    # Get the relative path inside the docs dir, without .rst, then
    # get the resource
    docname = str(source.relative_to(confdir)).split('.rst')[0]
    resource = resources.get(docname)

    if resource:
        for node in doctree.traverse(toctree):
            resource.toctree = [
                target for (flag, target) in node.attributes['entries']
            ]
            pass


@kb.event(SphinxEvent.DREAD, scope='articles')
def stamp_excerpt(kb_app: kb,
                  sphinx_app: Sphinx,
                  doctree: doctree):
    """ Walk the tree and extract excert into resource.excerpt """

    # First, find out which resource this is. Won't be easy.
    resources = sphinx_app.env.resources
    confdir = sphinx_app.confdir
    source = PurePath(doctree.attributes['source'])

    # Get the relative path inside the docs dir, without .rst, then
    # get the resource
    docname = str(source.relative_to(confdir)).split('.rst')[0]
    resource = resources.get(docname)

    if resource:
        # Stamp the excerpt on the resource
        excerpt = getattr(resource.props, 'excerpt', False)
        auto_excerpt = getattr(resource.props, 'auto_excerpt', False)
        if excerpt:
            resource.excerpt = excerpt
        elif not auto_excerpt:
            resource.excerpt = None
        else:
            # Extract the excerpt based on the number of paragraphs
            # in auto_excerpt
            resource.excerpt = get_rst_excerpt(doctree, auto_excerpt)


@kb.event(SphinxEvent.HPC, scope='articles')
def articles_into_html_context(
        kb_app: kb,
        sphinx_app: Sphinx,
        pagename,
        templatename: str,
        context,
        doctree: doctree) -> Dict[str, str]:
    resources = sphinx_app.env.resources
    navitems = [
        resource
        for resource in resources.values()
        if getattr(resource.props, 'in_nav', False)
    ]
    context['navmenu'] = sorted(navitems,
                                key=lambda x: (
                                    x.props.weight, attrgetter('title')(x))
                                )


@kb.dumper('toctrees')
def dump_settings(kb_app: kb, sphinx_env: BuildEnvironment):
    # First get the kb app configuration for widgets
    config = {
        k: v.__module__ + '.' + v.__name__
        for (k, v) in kb_app.config.toctrees.items()
    }

    toctrees = dict(
        config=config,
    )
    return dict(toctrees=toctrees)



"""

A UserDict container for resources.

This allows us to validate schemas on add plus provide some query-like
methods on the app.resources instance

"""
from collections import UserDict


class ResourcesContainer(UserDict):
    def __setitem__(self, name, value):
        self.data[name] = value



import datetime
import json

import os
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment import BuildEnvironment

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent
from kaybee.plugins.debugdumper.action import DumperAction


def datetime_handler(x):
    """ Allow serializing datetime objects to JSON """
    if isinstance(x, datetime.datetime):
        return x.isoformat()
    raise TypeError("Unknown type")


@kb.event(SphinxEvent.ECC, scope='debugdump', system_order=80)
def generate_debug_info(kb_app: kb, builder: StandaloneHTMLBuilder,
                        sphinx_env: BuildEnvironment):
    # If the config value doesn't enable dumping, bail out
    dumper_settings = sphinx_env.app.config.kaybee_settings.debugdumper
    use_debug = dumper_settings.use_debug

    if not use_debug:
        return

    # Get all the dumpers and dump their results
    dumpers = DumperAction.get_callbacks(kb_app)
    dumper_results = [dumper(kb_app, sphinx_env) for dumper in dumpers]
    result = {k: v for d in dumper_results for k, v in d.items()}

    # Now write the result to disk
    output_filename = os.path.join(builder.outdir, 'debug_dump.json')
    with open(output_filename, 'w') as f:
        json.dump(result, f, default=datetime_handler)



from typing import List, Any

from pydantic import BaseModel


class CorePropFilterModel(BaseModel):
    key: str
    value: Any

    class Config:
        ignore_extra = False


class BaseQueryModel(BaseModel):
    rtype: str = None
    limit: int = None
    parent_name: str = None
    sort_value: str = 'title'
    reverse: bool = False
    props: List[CorePropFilterModel] = []

    class Config:
        ignore_extra = False



from docutils.parsers.rst import Directive

from kaybee import app


class ResourceDirective(Directive):
    has_content = True

    @classmethod
    def get_resource_class(cls, resource_directive):
        """ Make this easy to mock """
        return app.kb.config.resources[resource_directive]

    @property
    def docname(self):
        return self.state.document.settings.env.docname

    @property
    def resources(self):
        return self.state.document.settings.env.resources

    def run(self):
        """ Run at parse time.

        When the documents are initially being scanned, this method runs
        and does two things: (a) creates an instance that is added to
        the site's widgets, and (b) leaves behind a placeholder docutils
        node that can later be processed after the docs are resolved.
        The latter needs enough information to retrieve the former.

        """

        rtype = self.name
        resource_content = '\n'.join(self.content)
        resource_class = ResourceDirective.get_resource_class(rtype)
        this_resource = resource_class(self.docname, rtype, resource_content)

        # Add this resource to the site
        self.resources[this_resource.docname] = this_resource

        # Don't need to return a resource "node", the document is the node
        return []



"""

Widget providing a number of queries to be performed then laid out
in columns.

"""
from typing import Dict

from kaybee.app import kb
from kaybee.plugins.queries.props_model import BaseQueryModel
from kaybee.plugins.queries.service import Query
from kaybee.plugins.widgets.base_widget import (
    BaseWidget,
    BaseWidgetModel,
)


class SectionqueryModel(BaseWidgetModel):
    query: BaseQueryModel


@kb.widget('sectionquery')
class SectionqueryWidget(BaseWidget):
    props: SectionqueryModel

    def make_context(self, context: Dict, sphinx_app):
        """ Put information into the context for rendering """

        query = self.props.query
        results = Query.filter_collection(
            sphinx_app.env.resources,
            rtype=query.rtype,
            parent_name=query.parent_name,
            sort_value=query.sort_value,
            limit=query.limit,
            reverse=query.reverse
        )
        context['results'] = results
        context['result_count'] = len(results)



from typing import Mapping

from pydantic import BaseModel
from sphinx.application import Sphinx
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.util.osutil import relative_uri

from kaybee.utils.models import load_model
from kaybee.utils.rst import rst_to_html


class BaseWidgetModel(BaseModel):
    name: str  # Must be unique on each page
    template: str = None  # Can come from class name

    class Config:
        ignore_extra = False


class BaseWidget:
    docname: str  # Widget instances get registered from a doc
    wtype: str  # This is the directive name, e.g. listing
    props: BaseWidgetModel
    rst_content: str  # The body rst converted to HTML

    def __init__(self, docname: str, wtype: str,
                 yaml_content: str,
                 rst_content: str = None):
        self.docname = docname
        self.wtype = wtype
        model = self.__annotations__['props']
        self.props: BaseWidgetModel = load_model(docname, model, yaml_content)

        if rst_content:
            self.content = rst_to_html(rst_content)

    def __repr__(self):
        return f'{self.docname}-{self.props.name}'

    def pathto_docname(self, target_docname):
        """ Mimic the sphinx.builders.html.pathto inline closure.

        We need a way to resolve relative URLs so hrefs work when a Sphinx
        site is served in a subdirectory. Sphinx has an odd inline function
        called pathto that has some local variables. This function serves the
        purpose for paths to docs as well as static resourcees. Let's split
        that and have our own, for widgets which aren't using the Sphinx
        Jinja2 renderer.

        """

        uri = relative_uri(self.docname, target_docname)
        return uri

    @property
    def template(self):
        """ Get the template from: YAML or class """

        # First try props
        if self.props.template:
            return self.props.template
        else:
            # Return the wtype of the widget, and we'll presume that,
            # like resources, there's a .html file in that directory
            return self.wtype

    def make_context(self, context: Mapping, sphinx_app: Sphinx):
        pass

    def render(self, sphinx_app: Sphinx, context):
        """ Given a Sphinx builder and context with sphinx_app in it,
         generate HTML """

        # Called from kaybee.plugins.widgets.handlers.render_widgets

        builder: StandaloneHTMLBuilder = sphinx_app.builder
        resource = sphinx_app.env.resources[self.docname]
        context['sphinx_app'] = sphinx_app
        context['widget'] = self
        context['resource'] = resource

        # make_context is optionally implemented on the concrete class
        # for each widget
        self.make_context(context, sphinx_app)

        # NOTE: Can use builder.templates.render_string
        template = self.template + '.html'
        html = builder.templates.render(template, context)
        return html

    def __json__(self):
        # The root has different rules about parents
        return dict(
            docname=self.docname,
            template=self.props.template,
            wtype=self.wtype,
            props=self.props.dict(),
            repr=repr(self),
        )



from typing import List

from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent


@kb.event(SphinxEvent.EBRD, scope='queries', system_order=40)
def initialize_query_service(kb_app: kb,
                             sphinx_app: Sphinx,
                             sphinx_env: BuildEnvironment,
                             docnames=List[str],
                             ):
    # TODO Doesn't look like this is needed
    sphinx_app.query = dict()



"""

Automatically re-build Sphinx docs while editing, serve docs on an
HTTP port, and also reload any browsers pointed to the docs.

"""

import glob
import sys
from pathlib import PurePosixPath

from livereload import Server, shell
from livereload.watcher import Watcher


class CustomWatcher(Watcher):
    """ Handle recursive globs with Python 3.5+ globbing  """

    paths = [
        'docs/**',
        'kaybee/**.py'
    ]
    root = 'docs/_build'

    def __init__(self):
        super().__init__()
        self.interp = sys.executable
        self.sphinx = str(PurePosixPath(self.interp).parent / 'sphinx-build')
        self.cmd = f'{self.interp} {self.sphinx} -E -b html docs docs/_build'

    def is_glob_changed(self, path, ignore=None):
        for f in glob.glob(path, recursive=True):
            if self.is_file_changed(f, ignore): # pragma: no cover
                return True
        return False


if __name__ == '__main__': # pragma: no cover
    watcher = CustomWatcher()
    server = Server(watcher=CustomWatcher())
    for p in watcher.paths:
        server.watch(p, shell(watcher.cmd),
                     ignore=lambda s: '_build' in s)
    server.serve(root=watcher.root, live_css=False)



"""

Managing the setting/retrieving of references via a UserDict.

Store/retrieve each as sphinx_app.env.references[rtype][label]

"""

from collections import UserDict
from typing import List, Mapping, Any

from kaybee.plugins.references.model_types import ReferencesType


class ReferencesContainer(UserDict):

    def get_reference(self, rtype: str, label: str):
        """ Return reference filed under rtype/label

         The references are organized by field/label, e.g. category/cat1.
         This lets us use a shorthand notation to go the resource, e.g.
         ref:category:cat1 instead of folder1/folder2/cat1.
         """

        # We are doing this instead of dictionary access in case we change
        # the storage later to a multidict thingy for optimization.

        reftype = self.data.get(rtype)
        if reftype:
            # The reftype might be "python" or "sphinx" or something else
            # from an Intersphinx registry, not something internal to
            # Kaybee.
            return reftype[label]

    def add_reference(self, reftype: str, label: str, target):
        """ Add reference object in references under rtype/label=target """

        # The self.data[reftype] dict springs into being during the
        # register_references event handler at startup, which looks in the
        # kb registry for all registered reference names.
        self.data[reftype][label] = target

    def resource_references(self, resource) -> Mapping[str, List[Any]]:
        """ Resolve and return reference resources pointed to by object

         Fields in resource.props can flag that they are references by
         using the references type. This method scans the model,
         finds any fields that are references, and returns the
         reference resources pointed to by those references.

         Note that we shouldn't get to the point of dangling references.
         Our custom Sphinx event should raise a references error
         during the build process (though maybe it is just a warning?)

         """

        references = dict()
        for reference_label in resource.props.references:
            references[reference_label] = []

            # Iterate over each value on this field, e.g.
            # tags: tag1, tag2, tag3
            for target_label in resource.props.references.get(reference_label):
                # Ask the site to get the object
                target = self.get_reference(reference_label, target_label)
                references[reference_label].append(target)

        return references



from kaybee.plugins.resources.base_resource import (
    BaseResource,
    BaseResourceModel
)


def is_reference_target(resource, rtype, label):
    """ Return true if the resource has this rtype with this label """

    prop = resource.props.references.get(rtype, False)
    if prop:
        return label in prop


class BaseReferenceModel(BaseResourceModel):
    label: str


class BaseReference(BaseResource):
    props: BaseReferenceModel
    is_reference = True

    def get_sources(self, resources):
        """ Filter resources based on which have this reference """

        rtype = self.rtype  # E.g. category
        label = self.props.label  # E.g. category1
        result = [
            resource
            for resource in resources.values()
            if is_reference_target(resource, rtype, label)
        ]
        return result



"""
New types that can be used in pydantic models.

"""


class ReferencesType(str):
    @classmethod
    def get_validators(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, list):
            raise ValueError(f'reference: expected list, not {type(v)}')
        for i in v:
            if not isinstance(i, str):
                prefix = 'reference: expected list of strings, '
                msg = f'{prefix} not {type(i)}'
                raise ValueError(msg)
        return v



import kaybee.plugins.articles
import kaybee.plugins.debugdumper.handlers
import kaybee.plugins.genericpage.handlers
import kaybee.plugins.localtemplates.handlers
import kaybee.plugins.queries.handlers
import kaybee.plugins.references.handlers
import kaybee.plugins.references.reference
import kaybee.plugins.resources.handlers
import kaybee.plugins.resources.resource
import kaybee.plugins.settings.handlers
import kaybee.plugins.sphinx_app.handlers
import kaybee.plugins.widgets.handlers
import kaybee.plugins.widgets.widget



"""

An out-of-the-box widget type.

This can be used in RST with something like::

   .. widget::
      name: someuniquename
      template: sometemplate.html

"""
from typing import Mapping

from sphinx.application import Sphinx

from kaybee.app import kb

from kaybee.plugins.widgets.base_widget import BaseWidget


@kb.widget('widget')
class Widget(BaseWidget):
    def make_context(self, context: Mapping, sphinx_app: Sphinx):
        pass



"""
pydantic field type for images.

Our resources have images associated with them. Not regular Sphinx images
which are part of the document. But rather, stuff associated with the
YAML, which we want to validate, post-process, and layout in our own
HTML.

"""
from pathlib import Path
import shutil

from pydantic import BaseModel
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.errors import SphinxError


class ImageModel(BaseModel):
    usage: str  # Multiple images can be attached so define usage
    filename: str

    def source_filename(self, docname: str, srcdir: str):
        """ Get the full filename to referenced image """

        docpath = Path(srcdir, docname)
        parent = docpath.parent
        imgpath = parent.joinpath(self.filename)

        # Does this exist?
        if not imgpath.exists():
            msg = f'Image does not exist at "{imgpath}"'
            raise SphinxError(msg)

        return imgpath

    def env_updated(self,
                    kb_app,
                    sphinx_app: Sphinx,
                    sphinx_env: BuildEnvironment,
                    resource
                    ):
        """ Make images and enter them in Sphinx's output writer """

        docname = resource.docname
        srcdir = sphinx_app.env.srcdir
        source_imgpath = self.source_filename(docname, srcdir)

        # Copy the image to the Sphinx build directory
        build_dir = sphinx_app.outdir
        docpath = Path(docname)
        parent = docpath.parent
        target_imgpath = str(Path(build_dir, parent, self.filename))
        shutil.copy(source_imgpath, target_imgpath)



from typing import List, Dict

from docutils import nodes
from docutils.readers import doctree
from sphinx.application import Sphinx
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment import BuildEnvironment

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent
from kaybee.plugins.references.container import ReferencesContainer


@kb.event(SphinxEvent.EBRD, scope='references', system_order=40)
def initialize_references_container(kb_app: kb,
                                    sphinx_app: Sphinx,
                                    sphinx_env: BuildEnvironment,
                                    docnames=List[str],
                                    ):
    if not hasattr(sphinx_app.env, 'references'):
        sphinx_app.env.references = ReferencesContainer()


@kb.event(SphinxEvent.EBRD, scope='references', system_order=50)
def register_references(kb_app: kb,
                        sphinx_app: Sphinx,
                        sphinx_env: BuildEnvironment,
                        docnames: List[str]):
    """ Walk the registry and add sphinx directives """

    references: ReferencesContainer = sphinx_app.env.references

    for name, klass in kb_app.config.resources.items():
        # Name is the value in the decorator and directive, e.g.
        # @kb.resource('category') means name=category
        if getattr(klass, 'is_reference', False):
            references[name] = dict()


@kb.event(SphinxEvent.ECC, scope='references', system_order=50)
def add_document_reference(kb_app: kb,
                           sphinx_builder: StandaloneHTMLBuilder,
                           sphinx_env: BuildEnvironment):
    # Moved from the resource directive .run(), this looks at each
    # resource.is_reference. If so, get the reference resource's label
    # and register it in sphinx_app.env.references
    sphinx_app: Sphinx = sphinx_env.app
    resources = sphinx_app.env.resources
    references = sphinx_app.env.references
    for resource in resources.values():
        if getattr(resource, 'is_reference', False):
            label = resource.props.label
            references.add_reference(resource.rtype, label, resource)


@kb.event(SphinxEvent.ECC, scope='references', system_order=60)
def validate_references(kb_app: kb,
                        sphinx_builder: StandaloneHTMLBuilder,
                        sphinx_env: BuildEnvironment):
    # Check the values in props to see if a resource has a reference
    # prop pointing at an invalid reference type, or an invalid reference
    # value. E.g.
    # category: cat1, cat2
    # ...where either 'category' isn't registered as a reference type or
    # 'cat1' isn't registered as a reference value for 'category'.
    resources = sphinx_env.resources
    references = sphinx_env.references

    for resource in resources.values():
        for (reference_label,
             reference_targets) in resource.props.references.items():
            # Make sure this label exists in site.reference
            try:
                site_reference = references[reference_label]
            except KeyError:
                msg = f'''\
            Document {resource.docname} uses unregistered reference label "{
            reference_label}"'''
                raise KeyError(msg)

            # Look through the values on this resource's
            # references['reference'] and ensure they are registered as a
            # target.
            for reference_target in reference_targets:
                try:
                    # Make sure references['reference'] has 'reference1'
                    assert reference_target in site_reference
                except AssertionError:
                    msg = f'''\
            Document {resource.docname} has "{reference_label}" with orphan {
        reference_target} '''
                    raise KeyError(msg)


@kb.event(SphinxEvent.MR, scope='references')
def missing_reference(kb_app: kb,
                      sphinx_app: Sphinx,
                      sphinx_env: BuildEnvironment,
                      node,
                      contnode):
    # The RST might have ref:somescheme values. Sphinx punts to a
    # callback such as this, which picks apart the scheme and expands it.
    references = sphinx_app.env.references
    refdoc = node['refdoc']
    try:
        target_rtype, target_label = node['reftarget'].split('-')
    except ValueError:
        # The scheme might be "python" or "sphinx" or something without
        # a dash. If so, just bail out here.
        return
    target = references.get_reference(target_rtype, target_label)

    if target:
        # Looks like this reference is using one of the Kaybee registered
        # reference types.
        if node['refexplicit']:
            # The ref has a title e.g. :ref:`Some Title <category-python>`
            dispname = contnode.children[0]
        else:
            # Use the title from the target
            dispname = target.title

        uri = sphinx_app.builder.get_relative_uri(refdoc, target.docname)
        newnode = nodes.reference('', '', internal=True, refuri=uri,
                                  reftitle=dispname)
        emp = nodes.emphasis()
        newnode.append(emp)
        emp.append(nodes.Text(dispname))
        return newnode


@kb.event(SphinxEvent.HPC, scope='references')
def references_into_html_context(
        kb_app: kb,
        sphinx_app: Sphinx,
        pagename,
        templatename: str,
        context,
        doctree: doctree) -> Dict[str, str]:
    # Always put the references db into Jinja2 context
    context['references'] = sphinx_app.env.references


@kb.dumper('references')
def dump_settings(kb_app: kb, sphinx_env: BuildEnvironment):
    # First get the kb app configuration for references
    config = {
        k: v.__module__ + '.' + v.__name__
        for (k, v) in kb_app.config.resources.items()
        if getattr(v, 'is_reference', False)
    }

    # Next, get the actual references in the app.references DB
    references = sphinx_env.references
    resources = sphinx_env.resources
    values = dict()
    for k, v in references.items():
        values[k] = dict()
        for label, resource in v.items():
            values[k][label] = resource.__json__(resources)

    references = dict(
        config=config,
        values=values
    )
    return dict(references=references)



"""

An out-of-the-box reference type.

This can be used in RST with something like::

   .. reference::
      template: sometemplate.html

"""

from kaybee.app import kb

from kaybee.plugins.references.base_reference import BaseReference


@kb.resource('reference')
class Reference(BaseReference):
    pass



"""

An action for the kb registry app.

"""

import dectate

from kaybee.plugins.genericpage.genericpage import Genericpage


class GenericpageAction(dectate.Action):
    config = {
        'genericpages': dict
    }

    def __init__(self, order: int = 40):
        super().__init__()
        self.order = order

    def identifier(self, genericpages):
        return str(self.order)

    # noinspection PyMethodOverriding
    def perform(self, obj, genericpages):
        genericpages[self.order] = obj

    @classmethod
    def get_genericpage(cls, kb_app):
        """ Return the one class if configured, otherwise default """

        # Presumes the registry has been committed
        q = dectate.Query('genericpage')
        klasses = sorted(q(kb_app), key=lambda args: args[0].order)
        if not klasses:
            # The site doesn't configure a genericpage,
            return Genericpage
        else:
            return klasses[0][1]



"""

A reference resource based on articles.

"""

from kaybee.plugins.references.base_reference import is_reference_target
from kaybee.plugins.articles.base_article import (
    BaseArticle,
    BaseArticleModel,
)


class BaseArticleReferenceModel(BaseArticleModel):
    label: str


class BaseArticleReference(BaseArticle):
    props: BaseArticleReferenceModel
    is_reference = True

    def get_sources(self, resources):
        """ Filter resources based on which have this reference """

        rtype = self.rtype  # E.g. category
        label = self.props.label  # E.g. category1
        result = [
            resource
            for resource in resources.values()
            if is_reference_target(resource, rtype, label)
        ]
        return result



from docutils.readers import doctree
from sphinx.application import Sphinx

from kaybee.app import kb
from kaybee.plugins.events import SphinxEvent


@kb.event(SphinxEvent.HPC, scope='sphinx_app')
def sphinx_app_html_context(
        kb_app: kb,
        sphinx_app: Sphinx,
        pagename,
        templatename: str,
        context,
        doctree: doctree,
):
    context['sphinx_app'] = sphinx_app
    context['kaybee_settings'] = sphinx_app.config.kaybee_settings



from kaybee.plugins.articles.base_article import (
    BaseArticle,
    BaseArticleModel
)


class BaseHomepageModel(BaseArticleModel):
    logo: str = None
    heading: str = None
    subheading: str = None
    hero_image: str = None


class BaseHomepage(BaseArticle):
    props: BaseHomepageModel



from docutils import nodes


class widget(nodes.General, nodes.Element):
    """ Generic invisible node that goes in doctree.

     When parsing doctrees, we might stumble across, in the middle
     of a document, a widget. The doctree needs a node that can
     be converted to a site.widgets[name] reference. Stick a
     generic, invisible (nothing rendered to output) node in the
     document, then extract (a) the kind of widget (classname) and
     (b) the identifier for the particular widget.

     """

    @property
    def name(self):
        """ This is the identifier for this node """

        return self['ids'][0]

    @property
    def rtype(self):
        """ The directive used, which finds the class needed

         If our RST has ``.. querylist::``
         """

        n = self.get('names', None)
        if n:
            return n[0]
        return None



from typing import Dict

from pydantic import ValidationError
from ruamel.yaml import load, Loader
from sphinx.errors import SphinxError


def load_model(docname: str, model, yaml_content: str) -> Dict:
    # If yaml_content is an empty string and parses to None, return
    # empty dic instead
    yaml_props = (load(yaml_content, Loader=Loader) or {})

    # Make the model, which validates, then do any extra validation
    try:
        m = model(**yaml_props)
    except ValidationError as exc:
        msg = f'Validation error in docname {docname}: {exc}'
        raise SphinxError(msg)
    return m



from typing import Mapping, List

from pydantic import BaseModel

from kaybee.utils.models import load_model


def parse_parent(docname):
    """ Given a docname path, pick apart and return name of parent """

    lineage = docname.split('/')
    lineage_count = len(lineage)

    if docname == 'index':
        # This is the top of the Sphinx project
        parent = None
    elif lineage_count == 1:
        # This is a non-index doc in root, e.g. about
        parent = 'index'
    elif lineage_count == 2 and lineage[-1] == 'index':
        # This is blog/index, parent is the root
        parent = 'index'
    elif lineage_count == 2:
        # This is blog/about
        parent = lineage[0] + '/index'
    elif lineage[-1] == 'index':
        # This is blog/sub/index
        parent = '/'.join(lineage[:-2]) + '/index'
    else:
        # This should be blog/sub/about
        parent = '/'.join(lineage[:-1]) + '/index'

    return parent  # This is a path


class BaseResourceModel(BaseModel):
    """ Kaybee default schema definitions for resources """

    template: str = None
    acquireds: Mapping[str, Mapping[str, str]] = None
    references: Mapping[str, List[str]] = dict()

    class Config:
        ignore_extra = False


class BaseResource:
    json_attrs = ('docname', 'rtype', 'parent')
    title: str = None  # Stamped on later by the handler
    props: BaseResourceModel

    def __init__(self, docname: str, rtype: str, yaml_content: str):
        self.docname = docname
        self.rtype = rtype
        self.parent = parse_parent(docname)

        model = self.__annotations__['props']
        self.props = load_model(docname, model, yaml_content)

    def __repr__(self):
        return self.docname

    def parents(self, resources):
        """ Split the path in name and get parents """

        if self.docname == 'index':
            # The root has no parents
            return []
        parents = []
        parent = resources.get(self.parent)
        while parent is not None:
            parents.append(parent)
            parent = resources.get(parent.parent)
        return parents

    def acquire(self, resources, prop_name):
        """ Starting with self, walk until you find prop or None """

        # Instance
        custom_prop = getattr(self.props, prop_name, None)
        if custom_prop:
            return custom_prop

        # Parents...can't use acquire as have to keep going on acquireds
        for parent in self.parents(resources):
            acquireds = parent.props.acquireds
            if acquireds:
                # First try in the per-type acquireds
                rtype_acquireds = acquireds.get(self.rtype)
                if rtype_acquireds:
                    prop_acquired = rtype_acquireds.get(prop_name)
                    if prop_acquired:
                        return prop_acquired

                # Next in the "all" section of acquireds
                all_acquireds = acquireds.get('all')
                if all_acquireds:
                    prop_acquired = all_acquireds.get(prop_name)
                    if prop_acquired:
                        return prop_acquired

        return

    def template(self, resources):
        """ Get the template from: YAML, hierarchy, or class """

        template_name = self.acquire(resources, 'template')
        if template_name:
            return template_name
        else:
            # We're putting an exception for "resource", the built-in
            # rtype/directive. We want it to work out-of-the-box without
            # requiring an _templates/resource.html in the docs project.
            # Instead, use the page.html the ships with Sphinx.
            if self.rtype == 'resource':
                return 'page'
            else:
                return self.rtype

    def __json__(self, resources):
        # The root has different rules about parents
        if self.docname == 'index':
            parent_docnames = self.parents(resources)
        else:
            parent_docnames = [p.docname for p in
                               self.parents(resources)]
        return dict(
            docname=self.docname,
            title=self.title,
            parent_docnames=parent_docnames,
            template=self.template(resources),
            rtype=self.rtype,
            parent=self.parent,
            props=self.props.dict(),
            repr=repr(self),
        )



from kaybee.plugins.articles import (
    article,
    featuretiles,
    handlers,
    homepage,
    querylist,
    category,
    section,
    sectionquery,
    sectionbox,
    toctree,
    videoplayer,
)



"""

An out-of-the-box homepage type.

This can be used in RST with something like::

   .. homepage::
      template: sometemplate.html

"""

from kaybee.app import kb

from kaybee.plugins.articles.base_homepage import BaseHomepage


@kb.resource('homepage')
class Homepage(BaseHomepage):
    pass



from sphinx.environment import BuildEnvironment

from kaybee.app import kb


@kb.dumper('settings')
def dump_settings(kb_app: kb, sphinx_env: BuildEnvironment):
    settings = sphinx_env.config.kaybee_settings.dict()
    return dict(settings=settings)



"""
Sphinx Template Bridge

Sphinx renders the HTML "body". We'd like to customize that without
writing a complete builder. Instead, we'll override the TemplateBridge
and take the HTML rendered output, then customize it before returning.

We'll do that by allowing the registration of "processors", for example
an XSLT transform.
"""

from sphinx.jinja2glue import BuiltinTemplateLoader

from kaybee.app import kb
from kaybee.plugins.postrenderer.action import PostrendererAction


class KaybeeBridge(BuiltinTemplateLoader):

    def get_kb(self):
        # Make it easy to mock this in text
        return kb

    def render(self, template, context):
        output = super().render(template, context)

        kb_app = self.get_kb()
        postrenderers = PostrendererAction.get_callbacks(kb_app)
        for postrenderer in postrenderers:
            pr = postrenderer()
            output = pr(output, context)

        return output



from datetime import datetime
from typing import List

from kaybee.plugins.articles.image_type import ImageModel
from kaybee.plugins.references.model_types import ReferencesType
from kaybee.plugins.resources.base_resource import (
    BaseResource,
    BaseResourceModel
)


class BaseArticleModel(BaseResourceModel):
    style: str = None
    css_class: str = None
    in_nav: bool = False
    nav_title: str = None
    weight: int = 0
    published: datetime = None
    category: ReferencesType = []
    tag: ReferencesType = []
    excerpt: str = None
    auto_excerpt: int = 1
    images: List[ImageModel] = []
    is_series: bool = False


class BaseArticle(BaseResource):
    props: BaseArticleModel
    excerpt: str = None  # Stamped on later by the handler
    toctree: List[str] = []  # Stamped on later by the handler

    def section(self, resources):
        """ Which section is this in, if any """

        section = [p for p in self.parents(resources) if p.rtype == 'section']
        if section:
            return section[0]
        return None

    def in_navitem(self, resources, nav_href):
        """ Given  href of nav item, determine if resource is in it """

        # The navhref might end with '/index' so remove it if so
        if nav_href.endswith('/index'):
            nav_href = nav_href[:-6]

        return self.docname.startswith(nav_href)

    @property
    def is_published(self):
        """ Return true if this resource has published date in the past """

        now = datetime.now()
        published = self.props.published
        if published:
            return published < now
        return False

    def series(self, resources):
        # Make sure the parent is a registered resource
        parent = resources.get(self.parent)
        if not parent:
            return None

        # The parent has to declare that it wants to have its children
        # in a "series", which is false by default.
        if not getattr(parent.props, 'is_series', False):
            return None

        results = []
        for docname in parent.toctree:
            resource = resources.get(docname)
            if resource:
                # We might have a non-resource page in the toctree,
                # so skip it if true
                excerpt = getattr(resource, 'excerpt', False)
                results.append(
                    dict(
                        docname=docname,
                        title=resource.title,
                        excerpt=excerpt,
                        current=self.docname == docname
                    )
                )
        return results

    def __json__(self, resources):
        d = super().__json__(resources)
        d['excerpt'] = self.excerpt
        d['section'] = getattr(self.section(resources), 'docname', '')
        d['toctree'] = self.toctree
        d['series'] = self.series(resources)

        return d
